#include <qcolor.h>
#include <qevent.h>
#include <qimage.h>

#include "../../../Core/Configuration/src/Configuration.h"
#include "../../../Utilities/src/Matrix.h"
#include "../../../Core/Interfaces/src/RendererInterface.h"
#include "../../../Core/Interfaces/src/FrameBufferInterface.h"
#include "EWAGLRendererConfiguration.h"
#include "EWAGLRenderer.h"

const GLfloat shiny_high[1]        = {100};  // a value between 1 & 128
const GLfloat shiny_low[1]         = {10};

#define NEAR_CLIP_PLANE 0.1
#define FAR_CLIP_PLANE 500000

#ifndef M_PI
	#define M_PI 3.14159265358979f
#endif


// **************
// public methods
// **************

EWAGLRenderer::EWAGLRenderer()
  : rendererWidget(0)
{
}

EWAGLRenderer::~EWAGLRenderer()
{
  if (rendererWidget != 0) {
    delete rendererWidget;
  }
}

void EWAGLRenderer::initialize (const bool isVisible, QWidget *parent)
{
  // delete the old widget, if necessary
  if (rendererWidget != 0) {
    delete rendererWidget;
  }

  // we allocate the renderer widget in any case, since the EWAGL rendering is implemented there
  rendererWidget = new EWAGLRendererWidget (isVisible, &rendererConfiguration, parent);
  // get informed whenever the rendererWidget is repainted
  this->connect (rendererWidget, SIGNAL (repainted()),
		 this, SLOT (handleRepainted()));
}

void EWAGLRenderer::setVisible (const bool enable)
{
  rendererWidget->setVisible (enable);
}

bool EWAGLRenderer::isVisible() const
{
  return rendererWidget->isVisible();
}

bool EWAGLRenderer::isInteractive() const
{
  return false;
}


bool EWAGLRenderer::providesPointPicking () const
{
  return false;
}

float EWAGLRenderer::getDepth (unsigned int x, unsigned int y) const
{
	GLint viewPort[4];
	glGetIntegerv(GL_VIEWPORT, viewPort);
	y = viewPort[3] - y - 1;
	GLfloat depth;
	glReadPixels(x, y, 1, 1, GL_DEPTH_COMPONENT, GL_FLOAT, &depth);
	return depth;
}


Vector3D EWAGLRenderer::getCameraSpaceNormal (unsigned int x, unsigned int y) const
{
  (void)x; (void)y;
  return Vector3D (0, 0, 1);
}



QWidget *EWAGLRenderer::getRendererWidget() const
{
  return rendererWidget;
}

RendererConfigurationInterface *EWAGLRenderer::getRendererConfiguration()
{
  return &rendererConfiguration;
}

QImage EWAGLRenderer::getRenderedImage()
{
  return rendererWidget->renderPixmap();
}

void EWAGLRenderer::setSceneView (const MyDataTypes::TransformationMatrix16f newSceneViewMatrix)
{
  rendererWidget->setSceneView (newSceneViewMatrix);
}

void EWAGLRenderer::getSceneView (MyDataTypes::TransformationMatrix16f sceneViewMatrix) const
{
  rendererWidget->getSceneView (sceneViewMatrix);
}

void EWAGLRenderer::setViewFrustum (const MyDataTypes::ViewFrustum newViewFrustum)
{
  rendererWidget->setViewFrustum (newViewFrustum);
}

MyDataTypes::ViewFrustum EWAGLRenderer::getViewFrustum() const
{
  return rendererWidget->getViewFrustum();
}

void EWAGLRenderer::setViewPortSize (const QSize newViewPortSize)
{
  rendererWidget->setViewPortSize (newViewPortSize);
}

QSize EWAGLRenderer::getViewPortSize() const
{
  return rendererWidget->getViewPortSize();
}

void EWAGLRenderer::setShadingEnabled (const bool enable) {
  rendererWidget->setShadingEnabled (enable);
}

bool EWAGLRenderer::isShadingEnabled() const
{
  return rendererWidget->isShadingEnabled();
}

void EWAGLRenderer::setLightDirection (const Vector3D newLightDirection)
{
  rendererWidget->setLightDirection (newLightDirection);
}

Vector3D EWAGLRenderer::getLightDirection() const
{
  return rendererWidget->getLightDirection();
}

void EWAGLRenderer::setBackgroundColor(QRgb newBackgroundColor) {
	rendererWidget->setBackgroundColor(newBackgroundColor);
}

QRgb EWAGLRenderer::getBackgroundColor() const {
	return rendererWidget->getBackgroundColor();
}

void EWAGLRenderer::setTwoSidedNormalsEnabled(const bool enable)
{
  (void)enable;
}

bool EWAGLRenderer::isTwoSidedNormalsEnabled() const
{
  return true;
}

void EWAGLRenderer::setShadowsEnabled(const bool enable)
{
  (void)enable;
}

bool EWAGLRenderer::isShadowsEnabled() const
{
  return false;
}


void EWAGLRenderer::toggleSelectionVisualization() {
	rendererWidget->setSelectionVisualizationEnabled(!rendererWidget->isSelectionVisualizationEnabled());
}

void EWAGLRenderer::setSelectionVisualizationEnabled (const bool enable)
{
  rendererWidget->setSelectionVisualizationEnabled(enable);
}

bool EWAGLRenderer::isSelectionVisualizationEnabled() const
{
  return rendererWidget->isSelectionVisualizationEnabled();
}

void EWAGLRenderer::setCutoffRadius (const float radius)
{
  rendererConfiguration.setCutOffRadius (radius);
}

float EWAGLRenderer::getCutoffRadius() const
{
  return rendererConfiguration.getCutOffRadius();
}

void EWAGLRenderer::setBlendingThresholds (const float constThreshold,
					    const float distThreshold,
					    const float angleThreshold)
{
  (void)constThreshold; (void)distThreshold; (void)angleThreshold;
}

void EWAGLRenderer::getBlendingThresholds (float *constThreshold, float *distThreshold, float *angleThreshold)
{
  // return default values
  *constThreshold = 1.0f;
  *distThreshold  = 1.0f;
  *angleThreshold = 1.0f;
}

void EWAGLRenderer::renderFrame (const int, const bool)
{
	rendererWidget->repaint();
}

void EWAGLRenderer::addToFrame (const std::vector<SurfelInterface*>* surfels, int nSurfels,
				 const int attributes, const bool updateRendererWidget)
{
  (void)surfels; (void)nSurfels; (void)attributes; (void)updateRendererWidget;
}

void EWAGLRenderer::subtractFromFrame (const std::vector<SurfelInterface*>* surfels, int nSurfels,
					const int attributes, const bool updateRendererWidget)
{
  (void)surfels; (void)nSurfels; (void)attributes; (void)updateRendererWidget;
}

FrameBufferInterface *EWAGLRenderer::getFrameBuffer() const
{
  // return dummy value (not supported)
  return 0;
}

// interaction stuff //////////////////////////////////////////////////////////

RendererInterface::InteractStateEnum  EWAGLRenderer::getModelInteractState(void) const
{
  return rendererWidget->getModelInteractState();
}

RendererInterface::InteractStateEnum  EWAGLRenderer::getViewpointInteractState(void) const
{
  return rendererWidget->getViewpointInteractState();
}

void  EWAGLRenderer::setInteractState(RendererInterface::InteractStateEnum state,
				                       RendererInterface::InteractStateAspectEnum target)
{
  rendererWidget->setInteractState(state, target);
}

// **********************************************************
  
// *************
// private slots
// *************

void EWAGLRenderer::handleRepainted()
{
  emit rendererWidgetRepainted();
}

// ****************
// access functions
// ****************

RendererInterface *createRenderer()
{
  return new EWAGLRenderer();
}

const char *getRendererType()
{
  return "EWAGL (Preview)";
}
